# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
from typing import Union, List
import math
import numbers
import functools
import operator

# External imports
import torch
from torch import Size
import torch.nn as nn
import torch.nn.init as init

# Local imports
import torchcvnn.nn.modules.batchnorm as bn

_shape_t = Union[int, List[int], Size]


class LayerNorm(nn.Module):
    r"""
    Implementation of the torch.nn.LayerNorm for complex numbers.

    Arguments:
        normalized_shape (int or list or torch.Size): input shape from an expected input of size $(*, normalized_shape[0], normalized_shape[1], ..., normalized_shape[-1])
        elementwise_affine (bool): a boolean value that when set to `True`, this module has learnable per-element affine parameters initialized to a diagonal matrix with diagonal element $1/\sqrt{2}$ (for weights) and zeros (for biases). Default: `True`
        bias (bool): if set to `False`, the layer will not learn an additive bias
    """

    def __init__(
        self,
        normalized_shape: _shape_t,
        elementwise_affine: bool = True,
        bias: bool = True,
        device: torch.device = None,
        dtype: torch.dtype = torch.complex64,
    ) -> None:
        super().__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        self.normalized_shape = tuple(normalized_shape)

        self.elementwise_affine = elementwise_affine

        self.combined_dimensions = functools.reduce(operator.mul, self.normalized_shape)
        if self.elementwise_affine:
            self.weight = torch.nn.parameter.Parameter(
                torch.empty((self.combined_dimensions, 2, 2), device=device)
            )
            if bias:
                self.bias = torch.nn.parameter.Parameter(
                    torch.empty((self.combined_dimensions,), device=device, dtype=dtype)
                )
            else:
                self.register_parameter("bias", None)
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        r"""
        Initialize the weight and bias. The weight is initialized to a diagonal
        matrix with diagonal $\frac{1}{\sqrt{2}}$.

        The bias is initialized to $0$.
        """
        with torch.no_grad():
            if self.elementwise_affine:
                # Initialize all the weights to zeros
                init.zeros_(self.weight)
                # And then fill in the diagonal with 1/sqrt(2)
                # w is C, 2, 2
                self.weight.view(-1, 2, 2)[:, 0, 0] = 1 / math.sqrt(2.0)
                self.weight.view(-1, 2, 2)[:, 1, 1] = 1 / math.sqrt(2.0)
                # Initialize all the biases to zero
                init.zeros_(self.bias)

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        """
        Performs the forward pass
        """
        # z: *, normalized_shape[0] , ..., normalized_shape[-1]
        z_ravel = z.view(-1, self.combined_dimensions).transpose(0, 1)

        # Compute the means
        mus = z_ravel.mean(axis=-1)  # normalized_shape[0]x normalized_shape[1], ...

        # Center the inputs
        z_centered = z_ravel - mus.unsqueeze(-1)
        z_centered = torch.view_as_real(
            z_centered
        )  # combined_dimensions,num_samples, 2

        # Transform the complex numbers as 2 reals to compute the variances and
        # covariances
        covs = bn.batch_cov(z_centered, centered=True)

        # Invert the covariance to scale
        invsqrt_covs = bn.inv_sqrt_2x2(covs)  # combined_dimensions, 2, 2
        # Note: the z_centered.transpose is to make
        # z_centered from (combined_dimensions, num_samples, 2) to (combined_dimensions, 2, num_samples)
        # So that the batch matrix multiply works as expected
        # where invsqrt_covs is (combined_dimensions, 2, 2)
        outz = torch.bmm(invsqrt_covs, z_centered.transpose(1, 2))
        outz = outz.contiguous()  # num_features, 2, BxHxW

        # Shift by beta and scale by gamma
        # weight is (num_features, 2, 2) real valued
        outz = torch.bmm(self.weight, outz)  # combined_dimensions, 2, num_samples
        outz = outz.transpose(1, 2).contiguous()  # combined_dimensions, num_samples, 2
        outz = torch.view_as_complex(outz)  # combined_dimensions, num_samples

        # bias is (C, ) complex dtype
        outz += self.bias.view(-1, 1)

        outz = outz.transpose(0, 1).contiguous()  # num_samples, comnbined_dimensions

        outz = outz.view(z.shape)

        return outz



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# External imports
import numpy as np

# Local imports
from . import parse_utils

descriptor_format = [
    ("file_id", 48, 16, "A", None),
    ("number_map_projection_record", 192, 6, "I", None),
]
descriptor_record_length = 720

summary_format = [
    ("scene_id", 20, 32, "A", None),
    ("number_sar_channels", 388, 4, "I", None),
    ("sensor_id", 412, 32, "A", None),
    ("line_content_indicator", 1670, 8, "A", None),  # RANGEbbb for L1.1 or OTHERbbb
    ("line_spacing", 1686, 16, "F16.7", None),
    ("pixel_spacing", 1702, 16, "F16.7", None),
    ("doppler_center_frequency_constant_term", 1734, 16, "F16.7", None),
    ("doppler_center_frequency_linear_term", 1750, 16, "F16.7", None),
    ("nominal_off_nadir", 1838, 16, "F16.7", None),
]
summary_record_length = 4096

map_projection_format = []
map_projection_record_length = 1620

platform_projection_format = []
platform_projection_record_length = 4680

altitude_data_format = [
    ("pitch", 40, 14, "E14.6", None),
    ("roll", 54, 14, "E14.6", None),
    ("yaw", 68, 14, "E14.6", None),
]
altitude_data_record_length = 16384

radiometric_data_format = [
    ("record_number", 0, 4, "B", None),
    ("calibration_factor", 20, 16, "F16.7", None),
]
radiometric_data_record_length = 9860

data_quality_format = []
data_quality_record_length = 1620

facility_format = []
facility_record_length = 325000 + 511000 + 3072 + 728000 + 5000


class LeaderFile:
    r"""
    Processing of the SAR trailer file.
    """

    def __init__(self, filepath):
        self.descriptor_record = {}
        self.summary_record = {}
        self.altitude_record = {}
        self.radiometric_record = {}
        with open(filepath, "rb") as fh:
            fh_offset = 0
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.descriptor_record,
                descriptor_format,
                1,
                descriptor_record_length,
                fh_offset,
            )
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.summary_record,
                summary_format,
                1,
                summary_record_length,
                fh_offset,
            )

            # Skip map projection record
            # Does not exist in L1.1
            if self.descriptor_record["number_map_projection_record"] == 1:
                fh_offset += map_projection_record_length

            # Skip platform projection record
            fh_offset += platform_projection_record_length

            # Altitude data record
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.altitude_record,
                altitude_data_format,
                1,
                altitude_data_record_length,
                fh_offset,
            )

            # Radiometric data record
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.radiometric_record,
                radiometric_data_format,
                1,
                radiometric_data_record_length,
                fh_offset,
            )

    @property
    def calibration_factor(self):
        cf = self.radiometric_record["calibration_factor"]
        return np.sqrt(10.0 ** ((cf - 32.0) / 10.0))

    def __repr__(self):
        descriptor_txt = parse_utils.format_dictionary(self.descriptor_record, 1)
        summary_txt = parse_utils.format_dictionary(self.summary_record, 1)
        altitude_txt = parse_utils.format_dictionary(self.altitude_record, 1)
        radiometric_txt = parse_utils.format_dictionary(self.radiometric_record, 1)
        return f"""
Descriptor:
{descriptor_txt}
Summary:
{summary_txt}
Altitude:
{altitude_txt}
Radiometric:
{radiometric_txt}
        """



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# External imports
import torch
import torch.nn as nn


class IndependentRealImag(nn.Module):
    """
    Generic module to apply a real valued activation function independently
    on both the real and imaginary part

    Arguments:
        fact: A nn.Module name of a real valued activation function
    """

    def __init__(self, fact: nn.Module):
        super().__init__()
        self.act_real = fact()
        self.act_imag = fact()

    def forward(self, z: torch.tensor) -> torch.tensor:
        """
        Performs the forward pass

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        return self.act_real(z.real) + self.act_imag(z.imag) * 1j


class CReLU(IndependentRealImag):
    """
    Applies a ReLU independently on both the real and imaginary parts

    $$
    CReLU(z) = ReLU(Re[z]) + ReLU(Im[z])j
    $$

    Only the quadrant where both `Re[z]` and `Im[z]` are negative is projected to
    $0$. Otherwise either the real and/or the imaginary part is preserved.

    """

    def __init__(self) -> None:
        super().__init__(nn.ReLU)


class CPReLU(IndependentRealImag):
    """
    Applies a PReLU independently on both the real and imaginary parts

    $$
    CPReLU(z) = PReLU(Re[z]) + PReLU(Im[z])j
    $$
    """

    def __init__(self) -> None:
        super().__init__(nn.PReLU)


class CELU(IndependentRealImag):
    """
    Applies a ELU independently on both the real and imaginary parts

    Not to confuse with `torch.nn.CELU`. For the complex equivalent of
    `torch.nn.CELU`, see [torchcvnn.nn.modules.activation.CCELU][]

    $$
    CELU(z) = ELU(Re[z]) + ELU(Im[z])j
    $$
    """

    def __init__(self) -> None:
        super().__init__(nn.ELU)


class CCELU(IndependentRealImag):
    """
    Applies a CELU independently on both the real and imaginary parts

    $$
    CCELU(z) = CELU(Re[z]) + CELU(Im[z])j
    $$
    """

    def __init__(self) -> None:
        super().__init__(nn.CELU)


class CGELU(IndependentRealImag):
    """
    Applies a GELU independently on both the real and imaginary parts

    $$
    CGELU(z) = GELU(Re[z]) + GELU(Im[z])j
    $$
    """

    def __init__(self) -> None:
        super().__init__(nn.GELU)


class CSigmoid(IndependentRealImag):
    """
    Applies a Sigmoid independently on both the real and imaginary parts

    as used in Nitta Tohru. An extension of the back-propagation algorithm to complex numbers. Neural Networks, 10(9):1391–1415, November 1997.


    $$
    CSigmoid(z) = Sigmoid(Re[z]) + Sigmoid(Im[z])j
    $$

    where the real valued sigmoid is applied in the right hand side terms.
    """

    def __init__(self) -> None:
        super().__init__(nn.Sigmoid)


class CTanh(IndependentRealImag):
    """
    Applies a Tanh independently on both the real and imaginary parts

    $$
    CTanh(z) = Tanh(Re[z]) + Tanh(Im[z])j
    $$

    where the real valued sigmoid is applied in the right hand side terms.
    """

    def __init__(self) -> None:
        super().__init__(nn.Tanh)


class zReLU(nn.Module):
    r"""
    Applies a zReLU

    $$
    zRELU(z) = \begin{cases} z & \mbox{if } Re[z] > 0 \mbox{ and } Im[z] > 0\\ 0 & \mbox{otherwise}  \end{cases}
    $$

    All the quadrant where both `Re[z]` and `Im[z]` are non negative are
    projected to $0$. In other words, only one quadrant is preserved.
    """

    def __init__(self):
        super().__init__()

    def forward(self, z: torch.Tensor):
        """
        Performs the forward pass.

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        pos_real = z.real > 0
        pos_img = z.imag > 0
        return z * pos_real * pos_img


class zAbsReLU(nn.Module):
    r"""
    Applies a zAbsReLU

    $$
    zAbsReLU(z) = \begin{cases} z & \mbox{if } |z| \geq a\\ 0 & \mbox{otherwise}  \end{cases}
    $$

    This cancels all the complex plane in the circle of radius $a$, where $a$ is
    trainable.
    """

    def __init__(self):
        super().__init__()
        self.a = torch.nn.parameter.Parameter(
            data=torch.Tensor([1.0]), requires_grad=True
        )

    def forward(self, z: torch.Tensor):
        """
        Performs the forward pass.

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        mask = z.abs() < self.a
        return z * mask


class zLeakyReLU(nn.Module):
    r"""
    Applies a zReLU

    $$
    zLeajyReLU(z) = \begin{cases} z & \mbox{if } Re[z] > 0 \mbox{ and } Im[z] >
    0\\ a.z & \mbox{otherwise}  \end{cases}
    $$

    """

    def __init__(self):
        super().__init__()
        self.a = torch.nn.parameter.Parameter(
            data=torch.Tensor([0.2]), requires_grad=True
        )

    def forward(self, z: torch.Tensor):
        """
        Performs the forward pass.

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        pos_real = z.real > 0
        pos_img = z.imag > 0
        return z * pos_real * pos_img + self.a * (z * ~(pos_real * pos_img))


class Mod(nn.Module):
    r"""
    Extracts the magnitude of the complex input: maps to $\mathbb{R}$

    $$
    Mod(z) = |z|
    $$

    This activation function allows to go from complex values to real
    values.

    """

    def __init__(self):
        super().__init__()

    def forward(self, z: torch.Tensor):
        """
        Performs the forward pass.

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        return torch.abs(z)


class modReLU(nn.Module):
    r"""
    Applies a ReLU with parametric offset on the amplitude, keeping the phase unchanged.

    $$
    modReLU(z = r e^{\theta j}) = ReLU(r + b) e^{\theta j})
    $$
    """

    def __init__(self):
        super().__init__()
        self.b = torch.nn.Parameter(torch.tensor(0.0, dtype=torch.float), True)

    def forward(self, z: torch.Tensor):
        """
        Performs the forward pass.

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        return nn.functional.relu(z.abs() + self.b) * torch.exp(1j * z.angle())


class Cardioid(nn.Module):
    r"""
    Applies a ReLU with parametric offset on the amplitude, keeping the phase unchanged.

    $$
    Cardioid(z = r e^{\theta j}) = \frac{1+\cos(\theta)}{2} z
    $$

    As proposed by Virtue et al. (2019). For real numbers, e.g. $\theta \in \{0,
    \pi\}$, it reduces to the ReLU :

    $$
    \forall r \in \mathbb{R}, \theta \in \{0, \pi\}, Cardiod(r e^{\theta j}) = Relu(r)
    $$
    """

    def __init__(self):
        super().__init__()
        self.b = torch.nn.Parameter(torch.tensor(0.0, dtype=torch.float), True)

    def forward(self, z: torch.Tensor):
        """
        Performs the forward pass.

        Arguments:
            z: the input tensor on which to apply the activation function
        """
        return 0.5 * (1 + torch.cos(z.angle())) * z



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import glob
import pathlib

# External imports
from torch.utils.data import Dataset
import numpy as np

# Local imports
from .slc_file import SLCFile


class SLCDataset(Dataset):
    r"""
    SLCDataset

    The format is described in https://uavsar.jpl.nasa.gov/science/documents/stack-format.html

    This object does not download the data for you, you must have the data on your local machine. For example, you can
    register and access data from the NASA JetLab https://uavsar.jpl.nasa.gov

    Note the datafiles can be quite large. For example, the quad polarization from Los Angeles SSurge_15305 is a bit
    more than 30 GB. If you take the downsampled datasets 2x8, it is 2GB.

    Note the 1x1 is 1.67 m slant range x 0.6 m azimuth.

    Note:
        As an example, using the example `read_slc.py`, with the SSurge_15305 stack provided
        by the UAVSar, the Pauli representation of the four polarizations is shown below :

        ![Pauli representation of a $3000 \times 3000$ crop of the SSurge_15305
        stack](../../../../../images/slc_SSurge_15305.png)

        The code may look like this :

        ```python
        import numpy as np
        import torchcvnn
        from torchcvnn.datasets.slc.dataset import SLCDataset

        def get_pauli(data):
            # Returns Pauli in (H, W, C)
            HH = data["HH"]
            HV = data["HV"]
            VH = data["VH"]
            VV = data["VV"]

            alpha = HH + VV
            beta = HH - VV
            gamma = HV + VH

            return np.stack([beta, gamma, alpha], axis=-1)


        patch_size = (3000, 3000)
        dataset = SLCDataset(
            rootdir,
            transform=get_pauli,
            patch_size=patch_size,
        )
        ```

    Arguments:
        rootdir: the path containing the SLC and ANN files
        transform : the transform applied to the patches. It applies
                    on a dictionnary of patches {'HH': np.array, 'HV': np.array, ...}
        patch_size: the dimensions of the patches to consider (rows, cols)
        patch_stride: the shift between two consecutive patches, default:patch_size
    """

    def __init__(
        self,
        rootdir: str = None,
        transform=None,
        patch_size: tuple = (128, 128),
        patch_stride: tuple = None,
    ):
        super().__init__()

        self.transform = transform
        self.patch_size = patch_size
        self.patch_stride = patch_stride
        if self.patch_stride is None:
            self.patch_stride = patch_size

        # Let us find all the SLC files
        # We group the polorizations of the stack together
        self.slcs = glob.glob(str(pathlib.Path(rootdir) / "*.slc"))
        self.slc_polarizations = {}
        self.patch_counts = {}
        for slc in self.slcs:
            slc_file = SLCFile(slc, patch_size=patch_size, patch_stride=patch_stride)
            slc_key = slc_file.key
            if slc_key not in self.slc_polarizations:
                self.slc_polarizations[slc_key] = {}

            self.slc_polarizations[slc_key][slc_file.polarization] = slc_file
            self.patch_counts[slc_key] = len(slc_file)

        # Sanity checks
        # 1- For every SLC stack, we must have the same number of patches
        # 2- All the SLCs must have the same number of polarizations
        polarization_count = None
        for slc_key, slc_polarizations in self.slc_polarizations.items():
            if polarization_count is None:
                polarization_count = len(slc_polarizations)
            else:
                assert polarization_count == len(slc_polarizations)

            patch_count = None
            for polarization, slc_file in slc_polarizations.items():
                if patch_count is None:
                    patch_count = len(slc_file)
                else:
                    assert patch_count == len(slc_file)

        self.nsamples = sum(self.patch_counts.values())

    def __len__(self):
        return self.nsamples

    def __getitem__(self, item):
        """
        Usefull params :
        - 1x1_slc_azimuth_pixel_spacing
        - 1x1_slc_range_pixel_spacing
        - global_average_squint_angle
        - center_wavelength
        - slc_SEGMENT_1x1_rows
        - slc_SEGMENT_1x1_columns
        """
        # Now, request the patch from the right SLC file
        assert 0 <= item < self.nsamples

        # 1- Find the right SLC file given self.patch_counts and item index
        for slc_key, count in self.patch_counts.items():
            if item < count:
                slcs = self.slc_polarizations[slc_key]
                break
            else:
                item -= count
        sorted_keys = sorted(slcs.keys())
        # 2- Find the right patch from all the polarizations
        patches = {pol: slcs[pol][item] for pol in sorted_keys}

        # 3a- Stack the patches
        # 3b- Apply the transform
        if self.transform is not None:
            patches = self.transform(patches)
        else:
            patches = np.stack([patchi for _, patchi in patches.items()])

        return patches



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# External imports
import torch

# Local imports
import torchcvnn.nn as c_nn


def test_crelu():
    activation = c_nn.CReLU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_cprelu():
    activation = c_nn.CPReLU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_celu():
    activation = c_nn.CELU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_ccelu():
    activation = c_nn.CCELU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_cgelu():
    activation = c_nn.CGELU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_csigmoid():
    activation = c_nn.CSigmoid()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_ctanh():
    activation = c_nn.CTanh()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_zrelu():
    activation = c_nn.zReLU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_zabsrelu():
    activation = c_nn.zAbsReLU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_zleakyrelu():
    activation = c_nn.zLeakyReLU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_mod():
    activation = c_nn.Mod()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_modReLU():
    activation = c_nn.modReLU()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


def test_cardioid():
    activation = c_nn.Cardioid()

    x = torch.randn((3, 3), dtype=torch.complex64)
    output = activation(x)


if __name__ == "__main__":
    test_crelu()
    test_cprelu()
    test_celu()
    test_ccelu()
    test_cgelu()
    test_csigmoid()
    test_ctanh()
    test_zrelu()
    test_zabsrelu()
    test_zleakyrelu()
    test_mod()
    test_modReLU()
    test_cardioid()



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def format_key_val(key_value):
    key, value = key_value
    return key.replace("_", " ").capitalize() + f" : {value}"


def format_dictionary(record, level=0):
    strs = map(format_key_val, record.items())
    # Add heading padding
    strs = map(lambda s: " " * (level * 4) + s, strs)
    return "\n".join(strs)


def read_field(fh, start_byte, num_bytes, type_bytes, expected):
    fh.seek(start_byte)
    data_bytes = fh.read(num_bytes)
    if type_bytes[0] == "A":
        value = data_bytes.decode("ascii")
    elif type_bytes[0] == "B":
        # Binary number representation, big_endian
        value = int.from_bytes(data_bytes, "big")
    elif type_bytes[0] == "I":
        data_str = data_bytes.decode("ascii").strip()
        if len(data_str) == 0:
            value = None
        else:
            value = int(data_str)
    elif type_bytes[0] == "E":
        # This is a  Em.n format
        # Exponential notation, right fill
        m, n = type_bytes[1:].split(".")
        value = float(data_bytes.decode("ascii").strip())
    elif type_bytes[0] == "F":
        # This is a  Fm.n format
        # Exponential notation, right fill
        m, n = type_bytes[1:].split(".")
        value = float(data_bytes.decode("ascii").strip())
    if expected is not None:
        assert value == expected

    return value


def parse_from_format(
    fh, obj, descriptor_format, num_records, record_length, base_offset
):
    if num_records == 1:
        for (
            field_name,
            start_byte,
            num_bytes,
            type_bytes,
            expected,
        ) in descriptor_format:
            value = read_field(
                fh, base_offset + start_byte, num_bytes, type_bytes, expected
            )
            obj[field_name] = value
        fh.seek(base_offset + record_length)
        return base_offset + record_length
    else:
        offset = base_offset
        for _ in range(num_records):
            record = {}
            for (
                field_name,
                start_byte,
                num_bytes,
                type_bytes,
                expected,
            ) in descriptor_format:
                value = read_field(
                    fh, offset + start_byte, num_bytes, type_bytes, expected
                )
                record[field_name] = value
            obj.append(record)
            offset += record_length
        fh.seek(offset)
        return offset



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
from typing import Optional

# External imports
import torch
import torch.nn as nn
from torch.nn.common_types import _size_2_t


class MaxPool2d(nn.Module):
    r"""Applies a 2D max pooling on the module of the input signal

    In the simplest case, the output value of the layer with input size $(N, C, H, W)$, output $(N, C, H_{out}, W_{out})$ and `kernel_size` `(kH, kW)` can be precisely described as:

    $$
    \begin{aligned}
    out(N_i, C_j, h, w) ={} & \max_{m=0, \ldots, kH-1} \max_{n=0, \ldots, kW-1} \\
                            & \text{|input|}(N_i, C_j, \text{stride[0]} \times h + m,
                                                   \text{stride[1]} \times w + n)
    \end{aligned}
    $$

    Internally, it is relying on the `torch.nn.MaxPool2d`
    
    Arguments:
        kernel_size: thr size of the window to take a max over
        stride: the stride of the window
        padding: implicit negative infinity padding to be added
        dilation: a parameter that controls the stride of elements in the window
        ceil_mode: when `True`, use `ceil` instead of `floor` to compute the output shape
        return_indices: if `True`, will return the max indices along with the outputs

    """

    def __init__(
        self,
        kernel_size: _size_2_t,
        stride: Optional[_size_2_t] = None,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        ceil_mode: bool = False,
        return_indices: bool = False,
    ) -> None:
        super().__init__()
        self.return_indices = return_indices
        self.m = nn.MaxPool2d(
            kernel_size,
            stride,
            padding,
            dilation,
            ceil_mode=ceil_mode,
            return_indices=True,
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        """
        Computes and return the MaxPool over the magnitude of the input
        """
        _, indices = self.m(torch.abs(z))

        if self.return_indices:
            return z.flatten()[indices], indices
        else:
            return z.flatten()[indices]


class AvgPool2d(nn.Module):
    """
    Implementation of torch.nn.AvgPool2d for complex numbers.
    Apply AvgPool2d on the real and imaginary part.
    Returns complex values associated to the AvgPool2d results.

    Arguments:
        kernel_size: thr size of the window to compute the average
        stride: the stride of the window
        padding: implicit negative infinity padding to be added
        ceil_mode: when `True`, use `ceil` instead of `floor` to compute the output shape
        count_include_pad: when `True`, will include the zero-padding in the averaging calculation
        divisor_override: if specified, it will be used as divisor, otherwise size of the pooling region will be used.
    """

    def __init__(
        self,
        kernel_size: _size_2_t,
        stride: Optional[_size_2_t] = None,
        padding: _size_2_t = 0,
        ceil_mode: bool = False,
        count_include_pad: bool = True,
        divisor_override: Optional[int] = None,
    ) -> None:
        super().__init__()
        if type(kernel_size) == int:
            self.kernel_size = [kernel_size] * 2 + [1]
        elif type(kernel_size) == tuple:
            if len(kernel_size) < 3:
                self.kernel_size = [kernel_size] + [1]
            else:
                self.kernel_size = kernel_size

        if type(stride) == int:
            self.stride = [stride] * 2 + [1]
        elif type(stride) == tuple:
            if len(stride) < 3:
                self.stride = [stride] + [1]
            else:
                self.stride = stride

        if type(padding) == int:
            self.padding = [padding] * 2 + [0]
        elif type(padding) == tuple:
            if len(padding) < 3:
                self.padding = [padding] + [0]
            else:
                self.padding = padding

        self.m = torch.nn.AvgPool3d(
            self.kernel_size,
            self.stride,
            self.padding,
            ceil_mode,
            count_include_pad,
            divisor_override,
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        """
        Computes the average over the real and imaginery parts.
        """
        return torch.view_as_complex(self.m(torch.view_as_real(z)))



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import math

# External imports
import torch
import torch.nn as nn
import torch.nn.init as init


def batch_cov(points: torch.Tensor, centered: bool = False) -> torch.Tensor:
    """
    Batched covariance computation
    Adapted from : https://stackoverflow.com/a/71357620/2164582

    Arguments:
        points: the (B, N, D) input tensor from which to compute B covariances
        centered: If `True`, assumes for every batch, the N vectors are centered.  default: `False`

    Returns:
        bcov: the covariances as a `(B, D, D)` tensor
    """
    B, N, D = points.size()
    if not centered:
        mean = points.mean(dim=1).unsqueeze(1)
        diffs = (points - mean).reshape(B * N, D)
    else:
        diffs = points.reshape(B * N, D)
    prods = torch.bmm(diffs.unsqueeze(2), diffs.unsqueeze(1)).reshape(B, N, D, D)
    bcov = prods.sum(dim=1) / (N - 1)  # Unbiased estimate
    return bcov  # (B, D, D)


def inv_2x2(M: torch.Tensor) -> torch.Tensor:
    r"""
    Computes the inverse of a tensor of shape [N, 2, 2].

    If we denote

    $$
    M = \begin{pmatrix} a & b \\ c & d \end{pmatrix}
    $$

    The inverse is given by

    $$
    M^{-1} = \frac{1}{Det M} Adj(M) = \frac{1}{ad - bc}\begin{pmatrix}d & -b \\ -c & a\end{pmatrix}
    $$

    Arguments:
        M: a batch of 2x2 tensors to invert, i.e. a $(B, 2, 2)$ tensor
    """
    det = torch.linalg.det(M).unsqueeze(-1).unsqueeze(-1)

    M_adj = M.clone()
    M_adj[:, 0, 0], M_adj[:, 1, 1] = M[:, 1, 1], M[:, 0, 0]
    M_adj[:, 0, 1] *= -1
    M_adj[:, 1, 0] *= -1
    M_inv = 1 / det * M_adj
    return M_inv


def sqrt_2x2(M: torch.Tensor) -> torch.Tensor:
    r"""
    Computes the square root of the tensor of shape [N, 2, 2].

    If we denote

    $$
    M = \begin{pmatrix} a & b \\ c & d \end{pmatrix}
    $$

    The square root is given by :

    $$
    \begin{align}
    \sqrt{M} &= \frac{1}{t} ( M + \sqrt{Det M} I)\\
    t &= \sqrt{Tr M + 2 \sqrt{Det M}}
    \end{align}
    $$

    Arguments:
        M: a batch of 2x2 tensors to invert, i.e. a $(B, 2, 2)$ tensor
    """
    N = M.shape[0]
    det = torch.linalg.det(M).unsqueeze(-1).unsqueeze(-1)
    sqrt_det = torch.sqrt(det)

    trace = torch.diagonal(M, dim1=-2, dim2=-1).sum(-1).unsqueeze(-1).unsqueeze(-1)
    t = torch.sqrt(trace + 2 * sqrt_det)

    sqrt_M = 1 / t * (M + sqrt_det * torch.eye(2, device=M.device).tile(N, 1, 1))
    return sqrt_M


def slow_inv_sqrt_2x2(M: torch.Tensor) -> torch.Tensor:
    """
    Computes the square root of the inverse of a tensor of shape [N, 2, 2]

    Arguments:
        M: a batch of 2x2 tensors to sqrt invert, i.e. a $(B, 2, 2)$ tensor
    """
    return sqrt_2x2(inv_2x2(M))


def inv_sqrt_2x2(M: torch.Tensor) -> torch.Tensor:
    """
    Computes the square root of the inverse of a tensor of shape [N, 2, 2]

    Arguments:
        M: a batch of 2x2 tensors to sqrt invert, i.e. a $(B, 2, 2)$ tensor
    """
    N = M.shape[0]
    det = torch.linalg.det(M).unsqueeze(-1).unsqueeze(-1)
    sqrt_det = torch.sqrt(det)

    trace = torch.diagonal(M, dim1=-2, dim2=-1).sum(-1).unsqueeze(-1).unsqueeze(-1)
    t = torch.sqrt(trace + 2 * sqrt_det)

    M_adj = M.clone()
    M_adj[:, 0, 0], M_adj[:, 1, 1] = M[:, 1, 1], M[:, 0, 0]
    M_adj[:, 0, 1] *= -1
    M_adj[:, 1, 0] *= -1
    M_sqrt_inv = (
        1 / t * (M_adj / sqrt_det + torch.eye(2, device=M.device).tile(N, 1, 1))
    )
    return M_sqrt_inv


class _BatchNormNd(nn.Module):
    r"""
    BatchNorm for complex valued neural networks. The same code applies for
    BatchNorm1d, BatchNorm2d, the only condition being the input tensor must be
    (batch_size, features, d1, d2, ..)

    The statistics will be computed over the $batch\_size \times d_1 \times d_2 \times ..$
    vectors of size $features$.

    As defined by Trabelsi et al. (2018)

    Arguments:
        num_features: $C$ from an expected input of size $(B, C)$
        eps: a value added to the denominator for numerical stability. Default $1e-5$.
        momentum: the value used for the running mean and running var computation. Can be set to `None` for cumulative moving average (i.e. simple average). Default: $0.1$
        affine: a boolean value that when set to `True`, this module has learnable affine parameters. Default: `True`
        track_running_stats: a boolean value that when set to `True`, this module tracks the running mean and variance, and when set to`False`, this module does not track such statistics, and initializes statistics buffers running_mean and running_var as None. When these buffers are None, this module always uses batch statistics. in both training and eval modes. Default: `True`
        cdtype: the dtype for complex numbers. Default torch.complex64
    """

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
        device: torch.device = None,
        cdtype: torch.dtype = torch.complex64,
    ) -> None:
        super().__init__()

        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats

        if self.affine:
            self.weight = torch.nn.parameter.Parameter(
                torch.empty((num_features, 2, 2), device=device)
            )
            self.bias = torch.nn.parameter.Parameter(
                torch.empty((num_features,), device=device, dtype=cdtype)
            )
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        if self.track_running_stats:
            # Register the running mean and running variance
            # These will not be returned by model.parameters(), hence
            # not updated by the optimizer although returned in the state_dict
            # and therefore stored as model's assets
            self.register_buffer(
                "running_mean",
                torch.zeros((num_features,), device=device, dtype=cdtype),
            )
            self.register_buffer(
                "running_var", torch.ones((num_features, 2, 2), device=device)
            )
            self.register_buffer(
                "num_batches_tracked", torch.tensor(0, dtype=torch.long, device=device)
            )
        else:
            self.register_buffer("running_mean", None)
            self.register_buffer("running_var", None)
            self.register_buffer(
                "num_batches_tracked",
                None,
            )
        self.reset_parameters()

    def reset_running_stats(self) -> None:
        if self.track_running_stats:
            self.running_mean.zero_()
            self.running_var.zero_()
            self.running_var[:, 0, 0] = 1 / math.sqrt(2.0)
            self.running_var[:, 1, 1] = 1 / math.sqrt(2.0)

    def reset_parameters(self) -> None:
        with torch.no_grad():
            self.reset_running_stats()
            if self.affine:
                # Initialize all the weights to zeros
                init.zeros_(self.weight)
                # And then fill in the diagonal with 1/sqrt(2)
                # w is C, 2, 2
                self.weight[:, 0, 0] = 1 / math.sqrt(2.0)
                self.weight[:, 1, 1] = 1 / math.sqrt(2.0)
                # Initialize all the biases to zero
                init.zeros_(self.bias)

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        # z : [B, C, d1, d2, ..] (complex)
        batch_size = z.shape[0]
        dim1 = z.shape[1]
        other_dims = z.shape[2:]

        xc = z.transpose(0, 1).reshape(self.num_features, -1)  # num_features, BxHxW

        if self.training or not self.track_running_stats:
            # For training
            # Or for testing but using the batch stats for centering/scaling

            # Compute the means
            mus = xc.mean(axis=-1)  # num_features means

            # Center the xc
            xc_centered = xc - mus.unsqueeze(-1)  # num_features, BxHxW
            xc_centered = torch.view_as_real(xc_centered)  # num_features, BxHxW, 2

            # Transform the complex numbers as 2 reals to compute the variances and
            # covariances
            covs = batch_cov(xc_centered, centered=True)  # 16 covariances matrices
        else:
            # The means come from the running stats
            mus = self.running_mean

            # Center the xc
            xc_centered = xc - mus.unsqueeze(-1)  # num_features, BxHxW
            xc_centered = torch.view_as_real(xc_centered)  # num_features, BxHxW, 2

            # The variance/covariance come from the running stats
            covs = self.running_var

        # Invert the covariance to scale
        invsqrt_covs = inv_sqrt_2x2(covs)  # num_features, 2, 2
        # Note: the xc_centered.transpose is to make
        # xc_centered from (C, BxHxW, 2) to (B, 2, BxHxW)
        # So that the batch matrix multiply works as expected
        # where invsqrt_covs is (C, 2, 2)
        outz = torch.bmm(invsqrt_covs, xc_centered.transpose(1, 2))
        outz = outz.contiguous()  # num_features, 2, BxHxW

        # Shift by beta and scale by gamma
        # weight is (num_features, 2, 2) real valued
        outz = torch.bmm(self.weight, outz)  # num_features, 2, BxHxW
        outz = outz.transpose(1, 2).contiguous()
        outz = torch.view_as_complex(outz)  # num_features, BxHxW

        # bias is (C, ) complex dtype
        outz += self.bias.view((self.num_features, 1))

        # With the following operation, weight
        # outz = outz.reshape(C, B, H, W).transpose(0, 1)
        outz = outz.reshape(dim1, batch_size, *other_dims).transpose(0, 1)

        if self.training and self.track_running_stats:
            self.running_mean = (
                1.0 - self.momentum
            ) * self.running_mean + self.momentum * mus
            if torch.isnan(self.running_mean).any():
                raise RuntimeError("Running mean divergence")

            self.running_var = (
                1.0 - self.momentum
            ) * self.running_var + self.momentum * covs
            if torch.isnan(self.running_var).any():
                raise RuntimeError("Running var divergence")
        return outz


class BatchNorm1d(_BatchNormNd):
    r"""
    BatchNorm for complex valued neural networks. The same code applies for
    BatchNorm1d, BatchNorm2d, the only condition being the input tensor must be
    (batch_size, features, d1, d2, ..)

    The statistics will be computed over the $batch\_size \times d_1 \times d_2 \times ..$
    vectors of size $features$.

    As defined by Trabelsi et al. (2018)

    Arguments:
        num_features: $C$ from an expected input of size $(B, C)$
        eps: a value added to the denominator for numerical stability. Default $1e-5$.
        momentum: the value used for the running mean and running var computation. Can be set to `None` for cumulative moving average (i.e. simple average). Default: $0.1$
        affine: a boolean value that when set to `True`, this module has learnable affine parameters. Default: `True`
        track_running_stats: a boolean value that when set to `True`, this module tracks the running mean and variance, and when set to`False`, this module does not track such statistics, and initializes statistics buffers running_mean and running_var as None. When these buffers are None, this module always uses batch statistics. in both training and eval modes. Default: `True`
        cdtype: the dtype for complex numbers. Default torch.complex64
    """
    pass


class BatchNorm2d(_BatchNormNd):
    r"""
    BatchNorm for complex valued neural networks. The same code applies for
    BatchNorm1d, BatchNorm2d, the only condition being the input tensor must be
    (batch_size, features, d1, d2, ..)

    The statistics will be computed over the $batch\_size \times d_1 \times d_2 \times ..$
    vectors of size $features$.

    As defined by Trabelsi et al. (2018)

    Arguments:
        num_features: $C$ from an expected input of size $(B, C)$
        eps: a value added to the denominator for numerical stability. Default $1e-5$.
        momentum: the value used for the running mean and running var computation. Can be set to `None` for cumulative moving average (i.e. simple average). Default: $0.1$
        affine: a boolean value that when set to `True`, this module has learnable affine parameters. Default: `True`
        track_running_stats: a boolean value that when set to `True`, this module tracks the running mean and variance, and when set to`False`, this module does not track such statistics, and initializes statistics buffers running_mean and running_var as None. When these buffers are None, this module always uses batch statistics. in both training and eval modes. Default: `True`
        cdtype: the dtype for complex numbers. Default torch.complex64
    """
    pass



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Standard imports
from typing import Optional

# External imports
import torch
import torch.nn as nn
from torch.nn.common_types import _size_any_t, _ratio_any_t


class Upsample(nn.Module):

    """
    Works by applying independently the same upsampling to both the real and
    imaginary parts.

    Note:
        With pytorch 2.1, applying the nn.Upsample to a complex valued tensor raises
        an exception "compute_indices_weights_nearest" not implemented for 'ComplexFloat'
        So it basically splits the input tensors in its real and imaginery
        parts, applies nn.Upsample on both components and view them as complex.


    Arguments:
        size (int or Tuple[int] or Tuple[int, int] or Tuple[int, int, int], optional):
            output spatial sizes
        scale_factor (float or Tuple[float] or Tuple[float, float] or Tuple[float, float, float], optional):
            multiplier for spatial size. Has to match input size if it is a tuple.
        mode (str, optional): the upsampling algorithm: one of ``'nearest'``,
            ``'linear'``, ``'bilinear'``, ``'bicubic'`` and ``'trilinear'``.
            Default: ``'nearest'``
        align_corners (bool, optional): if ``True``, the corner pixels of the input
            and output tensors are aligned, and thus preserving the values at
            those pixels. This only has effect when :attr:`mode` is
            ``'linear'``, ``'bilinear'``, ``'bicubic'``, or ``'trilinear'``.
            Default: ``False``
        recompute_scale_factor (bool, optional): recompute the scale_factor for use in the
            interpolation calculation. If `recompute_scale_factor` is ``True``, then
            `scale_factor` must be passed in and `scale_factor` is used to compute the
            output `size`. The computed output `size` will be used to infer new scales for
            the interpolation. Note that when `scale_factor` is floating-point, it may differ
            from the recomputed `scale_factor` due to rounding and precision issues.
            If `recompute_scale_factor` is ``False``, then `size` or `scale_factor` will
            be used directly for interpolation.
    """

    def __init__(
        self,
        size: Optional[_size_any_t] = None,
        scale_factor: Optional[_ratio_any_t] = None,
        mode: str = "nearest",
        align_corners: Optional[bool] = None,
        recompute_scale_factor: Optional[bool] = None,
    ) -> None:
        super().__init__()
        self.up_module = nn.Upsample(
            size, scale_factor, mode, align_corners, recompute_scale_factor
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        """
        Applies the forward pass
        """
        up_real = self.up_module(z.real).unsqueeze(-1)
        up_imag = self.up_module(z.imag).unsqueeze(-1)
        up_z = torch.cat((up_real, up_imag), axis=-1)
        return torch.view_as_complex(up_z)
